/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#include "dataserver.h"

namespace Conexus {

  DataServer::DataServer() :
      m_dispatcher( NULL ),
  m_use_dispatcher( false ) {
    m_running[ READ ] = false;
    m_running[ WRITE ] = false;
    for ( int i = READ; i <= WRITE; i++ ) {
      pthread_mutex_init( &m_mutex_queue[ i ], NULL );
      pthread_mutex_init( &m_mutex_can_stop[ i ], NULL );
      pthread_mutex_init( &m_mutex_can_start[ i ], NULL );
    }
  }


  DataServer::~DataServer() {
    if ( this->is_running() )
      stop();
  }

  bool DataServer::is_running() {
    return m_running[ READ ] | m_running[ WRITE ];
  }

  bool DataServer::is_read_running() {
    return m_running[ READ ];
  }

  bool DataServer::is_write_running() {
    return m_running[ WRITE ];
  }

  void DataServer::start( ReadWrite rw, bool use_dispatcher ) {
    // Start READ first then WRITE
    for ( int i = READ; i <= WRITE; i++ ) {
      if ( ! ( rw & i ) )
        continue;

      pthread_mutex_lock( &m_mutex_can_start[ i ] );

      if ( m_running[ i ] ) {
        pthread_mutex_unlock( &m_mutex_can_start[ i ] );
        continue;
      }

      if ( i == READ ) {
        m_use_dispatcher = use_dispatcher;
        if ( m_use_dispatcher ) {
          m_dispatcher = new Glib::Dispatcher();
          if ( m_dispatcher && ! m_dispatcher_connection )
            m_dispatcher_connection = m_dispatcher->connect( sigc::mem_fun( *this, &DataServer::emit_received_data ) );
        }
      }

      m_running[ i ] = true;
      pthread_attr_init( &m_thread_attr[ i ] );

      if ( i == READ )
        pthread_create( &m_thread[ i ], &m_thread_attr[ i ], &DataServer::read_thread_main_static, this );
      else
        pthread_create( &m_thread[ i ], &m_thread_attr[ i ], &DataServer::write_thread_main_static, this );

      pthread_mutex_unlock( &m_mutex_can_start[ i ] );
    }
  }

  void DataServer::stop( ReadWrite rw ) {
    // Stop READ first then WRITE
    for ( int i = READ; i <= WRITE; i++ )
      if ( ( rw & i ) && m_running[ i ] ) {
        pthread_mutex_lock( &m_mutex_can_stop[ i ] );  // Grab a stop lock before issuing pthread_cancel()
        pthread_mutex_lock( &m_mutex_can_start[ i ] ); // Grab a start lock to prevent starting before cleanup is done
        pthread_cancel( m_thread[ i ] ); // Cancel the running thread to shut down
        pthread_join( m_thread[ i ], NULL ); // Wait for the cancel to take effect before unlocking
        pthread_mutex_unlock( &m_mutex_can_stop[ i ] );
        m_running[ i ] = false; // We're stopped and can't start
        if ( i == READ )
          this->read_thread_cleanup();
        else
          this->write_thread_cleanup();
        pthread_mutex_unlock( &m_mutex_can_start[ i ] ); // Now that cleanup is done release the start lock
      }
  }

  void DataServer::read_thread_cleanup() {
    if ( m_dispatcher_connection )
      m_dispatcher_connection.disconnect();
    if ( m_dispatcher ) {
      m_use_dispatcher = false;
      delete m_dispatcher;
      m_dispatcher = NULL;
    }
  }

  void DataServer::write_thread_cleanup() {}

  sigc::signal<void, Data> DataServer::signal_data() {
    return m_signal_data;
  }

  sigc::signal<void, CData> DataServer::signal_cdata() {
    return m_signal_cdata;
  }

  void DataServer::read_thread_main() {}

  void DataServer::write_thread_main() {}

  void DataServer::queue_received_data( Data d ) {
    pthread_mutex_lock( &m_mutex_queue[ READ ] );
    m_read_queue.push( d );
    pthread_mutex_unlock( &m_mutex_queue[ READ ] );
    if ( m_use_dispatcher )
      m_dispatcher->emit();
    else
      this->emit_received_data();
  }

  void DataServer::queue_write_data( Data d ) {
    pthread_mutex_lock( &m_mutex_queue[ WRITE ] );
    m_write_queue.push( d );
    pthread_mutex_unlock( &m_mutex_queue[ WRITE ] );
  }

  void DataServer::emit_received_data( ) {
    Data current;
    pthread_mutex_lock( &m_mutex_queue[ READ ] );
    while ( ! m_read_queue.empty() ) {
      current = m_read_queue.front();
      m_read_queue.pop();
      pthread_mutex_unlock( &m_mutex_queue[ READ ] );
      if ( m_signal_cdata.slots().begin() != m_signal_cdata.slots().end() )
        m_signal_cdata.emit( CData( current.data.get(), current.size ) );
      sigc::signal<void, Data>::iterator next;
      next = m_signal_data.slots().begin();
      next++;
      for ( sigc::signal<void, Data>::iterator i = m_signal_data.slots().begin(); i != m_signal_data.slots().end(); i++ ) {
        if ( next != m_signal_data.slots().end() ) {
          ( *i ) ( Data( current.data.get(), current.size ) );
          next++;
        } else {
          ( *i ) ( current );
          current.clear();
        }
      }
      pthread_mutex_lock( &m_mutex_queue[ READ ] );
    }
    pthread_mutex_unlock( &m_mutex_queue[ READ ] );
  }

  sigc::signal< void, size_t > DataServer::signal_data_received( ) {
    return m_signal_data_received;
  }

  sigc::signal< void, size_t > DataServer::signal_data_transmitted( ) {
    return m_signal_data_transmitted;
  }

  void* DataServer::read_thread_main_static( void * p ) {
    DataServer * s = ( DataServer* ) p;
    s->read_thread_main();
    return p;
  }

  void* DataServer::write_thread_main_static( void * p ) {
    DataServer * s = ( DataServer* ) p;
    s->write_thread_main();
    return p;
  }

}
