/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "endpoint.h"

namespace Conexus
  {

  Endpoint::Endpoint(bool close_on_destruction):
      m_close_on_destruction(close_on_destruction),
      m_state(CLOSED),
      m_readable(true),
      m_writable(true),
      m_read_blocked(false),
      m_write_blocked(false)
  {}


  Endpoint::~Endpoint()
  {}

  ssize_t Endpoint::write(Data data, IOMETHOD block)
  {
    return write(data.data.get(), data.size);
  }

  ssize_t Endpoint::write(CData data, IOMETHOD block)
  {
    return write(data.data.get(), data.size);
  }

  void Endpoint::change_state( unsigned long states ) throw (state_error)
  {
    if (states == UNCHANGED)
      return;

    // check for inconsistent states
    if ( states&CLOSED && states&OPENED)
      throw error::state::inconsistent();

    // perform the proper actions
    if ( is_open() && states&CLOSED )
      {
        close();
        if ( is_open() )
          throw error::state::failed();
      }
    else if ( is_closed() && states&OPENED )
      {
        open();
        if ( is_closed() )
          throw error::state::failed();
      }

  }

  void Endpoint::set_state_opened()
  {
    if (is_open())
      return;
    m_state &= ~CLOSED;
    m_state |= OPENED;
    m_signal_opened.emit();
  }

  void Endpoint::set_state_closed()
  {
    if (is_closed())
      return;
    m_state &= ~OPENED;
    m_state |= CLOSED;
    m_signal_closed.emit();
  }

  void Endpoint::block_read(bool read_block)
  {
    m_read_blocked = read_block;
    m_signal_read_write_block_changed.emit(m_read_blocked, m_write_blocked);
  }

  void Endpoint::block_write(bool write_block)
  {
    m_write_blocked = write_block;
    m_signal_read_write_block_changed.emit(m_read_blocked, m_write_blocked);
  }

  void Endpoint::block_read_write(bool read_block, bool write_block)
  {
    m_read_blocked = read_block;
    m_write_blocked = write_block;
    m_signal_read_write_block_changed.emit(m_read_blocked, m_write_blocked);
  }

  bool Endpoint::close_on_destruction() const
    {
      return m_close_on_destruction;
    }


  void Endpoint::set_close_on_destruction(bool value)
  {
    m_close_on_destruction = value;
  }

  void Endpoint::close_and_reopen( unsigned long new_state )
  {
    unsigned long prev_state = state();
    change_state(CLOSED);
    if (new_state == UNCHANGED)
      change_state(prev_state);
    else
      change_state(new_state);
  }

  sigc::signal<void> Endpoint::signal_opened()
  {
    return m_signal_opened;
  }

  sigc::signal<void> Endpoint::signal_closed()
  {
    return m_signal_closed;
  }

  sigc::signal<void, bool, bool> Endpoint::signal_read_write_block_changed()
  {
    return m_signal_read_write_block_changed;
  }

  bool Endpoint::is_open()
  {
    return m_state&OPENED;
  }

  bool Endpoint::is_closed()
  {
    return m_state&CLOSED;
  }

  bool Endpoint::is_read_blocked()
  {
    return m_read_blocked;
  }

  bool Endpoint::is_write_blocked()
  {
    return m_write_blocked;
  }

  unsigned long Endpoint::state()
  {
    return m_state;
  }


} // namespace Conexus

Conexus::Endpoint& operator<<(Conexus::Endpoint& io, Conexus::Data d)
{
  io.write(d);
  return io;
}

Conexus::Endpoint& operator<<(Conexus::Endpoint& io, Conexus::CData d)
{
  io.write(d);
  return io;
}

Conexus::Endpoint& operator>>(Conexus::Endpoint& io, Conexus::Data d)
{
  d = io.read();
  return io;
}

