/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef CONEXUSADDRESS_ERROR_H
#define CONEXUSADDRESS_ERROR_H

#include <conexus/error.h>
#include <netdb.h>

namespace Conexus
  {

    /**
     * Base class for all address errors.
     * @author Rick L Vinyard Jr
     * @ingroup errors
     */
    class address_error: public conexus_error
      {
      public:
      address_error(int e=0): conexus_error("Unknown address error.", e,"address")
        { }
      address_error(const std::string s, int e=0): conexus_error(s, e,"address")
        { }
      }
    ;

  namespace error
  {

  namespace address
      {

      /**
       * Thrown when trying to create an address from a string that is improperly formed.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
      class bad_string: public address_error
        {
        public:
          bad_string(): address_error("Address string is improperly formed.")
          { }
        }
      ;

      /**
       * Thrown when a netmask is required and the string or numeric is not a valid netmask.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class bad_netmask: public address_error
       {
          public:
          bad_netmask(): address_error("The provided value is not a valid netmask."){ }
       };

      /**
       * EAI_AGAIN
       * The name could not be resolved at this time. Try again later.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class again: public address_error
       {
         public:
           again(): address_error("The name could not be resolved at this time. Try again later.", EAI_AGAIN) { }
       };

      /**
       * EAI_BADFLAGS
       * The flags parameter has an invalid value.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class bad_flags: public address_error
       {
         public:
           bad_flags(): address_error("The flags parameter has an invalid value.", EAI_BADFLAGS) { }
       };

      /**
       * EAI_FAIL
       * A non-recoverable error occurred.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class fail: public address_error
       {
         public:
           fail(): address_error("A non-recoverable error occurred.", EAI_FAIL) { }
       };

      /**
       * EAI_MEMORY
       * Out of memory.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class memory: public address_error
       {
         public:
           memory(): address_error("Out of memory.", EAI_MEMORY) { }
       };

       
      /**
       * EAI_NONAME
       * The  name  does  not  resolve for the supplied parameters.
       * NI_NAMEREQD is set and the host’s name cannot be located,
       * or neither hostname nor service name were requested.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class no_name: public address_error
       {
         public:
           no_name(): address_error("The  name  does  not  resolve for the supplied parameters.", EAI_NONAME) { }
       };

      /**
       * EAI_SYSTEM
       * A system error occurred.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class system: public address_error
       {
         public:
           system(): address_error("A system error occurred.", EAI_SYSTEM) { }
       };

      /**
       * EAI_FAMILY:
       * The requested address family is not supported at all.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class family: public address_error
       {
         public:
           family(): address_error("The requested address family is not supported at all.", EAI_FAMILY) { }
       };

      /**
       * EAI_SOCKTYPE:
       * The requested socket type is not supported at all.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class socket_type: public address_error
       {
         public:
           socket_type(): address_error("The requested socket type is not supported at all.", EAI_SOCKTYPE) { }
       };

      /**
       * EAI_SERVICE:
       * The requested service is not available for the requested socket type.
       * It may be available through another socket type.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class service: public address_error
       {
         public:
           service(): address_error("The requested service is not available for the requested socket type.", EAI_SERVICE) { }
       };

      /**
       * EAI_ADDRFAMILY:
       * The specified network host does not have any network addresses in the requested address family.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class address_family: public address_error
       {
         public:
           address_family(): address_error("The specified network host does not have any network addresses in the requested address family.", EAI_ADDRFAMILY) { }
       };


      /**
       * EAI_NODATA:
       * The specified network host exists, but does not have any network addresses defined.
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class no_data: public address_error
       {
         public:
           no_data(): address_error("The specified network host exists, but does not have any network addresses defined.", EAI_NODATA) { }
       };

      /**
       * Trying to convert an IPv6 address to IPv4 that is not IPv4 compatible
       * @author Rick L Vinyard Jr
       * @ingroup errors
       */
       class ipv6_not_ipv4_compatible: public address_error
       {
         public:
           ipv6_not_ipv4_compatible(): address_error("Converting IPv6 address to IPv4 when address is not IPv4 compatible.") { }
       };


      }

  }

}

#endif // CONEXUSBIND_ERROR_H
