  /***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef CONEXUS_IPADDRESS_H
#define CONEXUS_IPADDRESS_H

#include <string>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sigc++/sigc++.h>

#include <conexus/address.h>
#include <conexus/error.h>

namespace Conexus
  {

    typedef enum IP_ADDRESS_PORT { IP_ADDRESS=0x01, IP_PORT=0x02 } IP_ADDRESS_PORT;

  /**
     * @namespace conexus::IPBase
   * The conexus::IP namespace contains classes capable of communicating at
   * OSI Layers 3 (Network) and 4 (Transport) which includes support for
   * raw IP, TCP/IP and UDP/IP communications [ For more information see IP(7) ]
   */
  namespace IPBase
    {

    /**
     * IP network address interface.
     *
     * This class represents an IP address string based interface to both IPv4 and IPv6 addresses
     * which are further implemented in IPv4::Address and IPv6::Address.
     *
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     * @ingroup IP
     */
    class AddressBase: public Conexus::Address
      {
      public:
        typedef boost::shared_ptr<AddressBase> pointer;
        
        AddressBase() { }

        AddressBase(const struct sockaddr_storage& addr): Conexus::Address(addr) { }

        virtual ~AddressBase() { }

        /** Returns the address as a string. */
        virtual std::string address_string() const = 0;

        /** Returns the netmask associated with this address as a string. */
        virtual std::string subnet_mask_string() const = 0;

        /** Returns the prefix length of the subnet mask associated with this address. */
        virtual unsigned prefix_length() const = 0;

        /** Returns the prefix length of the subnet mask associated with this address as a string. */
        virtual std::string prefix_length_string() const = 0;

        /** Returns the address as an address/subnetmask string. */
        virtual std::string address_subnet_mask_string() const = 0;

        /** Returns the address as a CIDR address/prefixlen string. */
        virtual std::string cidr_address_string() const = 0;

        /** Returns the prefix part of this address as a string. */
        virtual std::string prefix_string() const = 0;

        /** Returns the local part of this address as a string. */
        virtual std::string local_address_string() const = 0;

        /** Returns the port number associated with this address. */
        virtual uint16_t port() const = 0;

        /** Returns the port number associated with this address as a string. */
        virtual std::string port_string() const = 0;

        /** Sets the address to the specified value. */
        virtual void set_address(std::string address) = 0;

        /** Sets the address and port to the specified value. */
        virtual void set_address(std::string address, uint16_t port) = 0;

        /** Sets the prefix length to the specified value. */
        virtual void set_prefix_length(unsigned prefix_length) = 0;

        /** Sets the prefix length to the specified value. */
        virtual void set_prefix_length(std::string prefix_length) = 0;

        /** Sets the port portion of this address. */
        virtual void set_port(uint16_t port) = 0;

        /** Sets the port to the specified value. */
        virtual void set_port(std::string port) = 0;

        /** Performs a DNS query on the currently set address and returns the hostname string. */
        virtual std::string hostname() const = 0;

        /** Performs a service name lookup on the currently set port and returns the servicename string */
        virtual std::string servicename() const = 0;

        /** True if this is a multicast address */
        virtual bool is_multicast() const = 0;

        /** True if this is the any address */
        virtual bool is_any() const = 0;

        /** True if this is the loopback address */
        virtual bool is_loopback() const = 0;

        sigc::signal<void, unsigned> signal_address_port_changed( )  { return m_signal_address_port_changed; }
        sigc::signal<void> signal_prefix_changed( ) { return m_signal_prefix_changed; }
        
        virtual const std::string& object_type() { static std::string s("Conexus::IP::Address"); return s; }
        
        virtual bool is_ipv4() = 0;
        
        virtual bool is_ipv6() = 0;

        protected:
          sigc::signal<void, unsigned> m_signal_address_port_changed;
          sigc::signal<void> m_signal_prefix_changed;
          
      };

  }

}

#endif
