/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#include "ipv4_tcp.h"

#include <iostream>

#include <sys/ioctl.h>
#include <linux/sockios.h>

namespace Conexus
{
  namespace IPv4
  {

    TCP::TCP(uint16_t port):
        IP(SOCK_STREAM, IPPROTO_TCP)
    {
      m_remote_address.set_port(port);
    }


    TCP::TCP(int accepted_fd, bool):
        IP(SOCK_STREAM, IPPROTO_TCP)
        {
          m_fd = accepted_fd;
          m_state |= BOUND | CONNECTED | ACCEPTED;
        }

        TCP::pointer TCP::create(uint16_t port) {
          return pointer( new TCP(port) );
        }
        
        TCP::pointer TCP::create( int accepted_fd, bool b) {
          return pointer( new TCP(accepted_fd, b) );
        }

    TCP::~TCP() throw()
    {
    }

    int TCP::pending_input_size()
    {
      int rsize;
      int error;
      error = ioctl(m_fd, SIOCINQ, &rsize);
      return rsize;
    }

    int TCP::pending_output_size()
    {
      int rsize;
      int error;
      error = ioctl(m_fd, SIOCOUTQ, &rsize);
      return rsize;
    }

    ssize_t TCP::write(const void* data, size_t size) throw (write_error)
    {
        if ( ! is_accepted() && ! is_connected() )
          try
        {
          change_state(CONNECTED);
        }
        catch (error::state::failed)
        {
          throw error::write::not_connected();
        }

        return Socket::write(data, size);
    }

    Data TCP::read(size_t s) throw (read_error)
    {
      if (! ( is_accepted() || is_bound() ) )
        try
      {
        change_state(BOUND);
      }
      catch (error::state::failed)
      {
        throw error::read::not_bound();
      }

      return Socket::read(pending_input_size());
    }

    void TCP::connect( ) throw( connect_error )
    {
      if (m_remote_address_set)
        Socket::connect(m_remote_address);
      else
        throw error::connect::no_remote_address();
    }

    void TCP::on_local_interface_changed(unsigned which)
    {
      // if this is an accepted socket, there is nothing to do when the local interface is changed
      if ( is_accepted() )
        return;

      // but, if this is a client side socket, and the socket is already bound, close and reopen to
      // same state with new local port and interface
      if ( is_bound() )
        close_and_reopen();
    }

    void TCP::on_remote_address_changed(unsigned which)
    {
      // if this is an accepted socket, there is nothing to do when the remote address is changed
      if ( is_accepted() )
        return;

      // but, if this is a client side socket, and the socket is already connected, close and reopen to
      // same state with new local port and interface
      if ( is_connected() )
        close_and_reopen();
    }

  } // namespace IPv4
} // namespace conexus
