/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "ipv4_udp.h"

#include <iostream>

#include <sys/ioctl.h>
#include <linux/sockios.h>

namespace Conexus
  {
  namespace IPv4
    {

    UDP::UDP(uint16_t localport, uint16_t remoteport):
        IP(SOCK_DGRAM, IPPROTO_UDP),
        m_write_without_connect(false),
        m_multicast_auto_join(true),
        m_multicast_request(NULL)
    {
      m_local_interface.set_port(localport);
      m_remote_address.set_port(remoteport);
    }

    UDP::pointer UDP::create(uint16_t localport, uint16_t remoteport) {
      return pointer( new UDP( localport, remoteport ) );
    }

    UDP::~UDP() throw()
    {
      if (m_multicast_request)
        leave();
    }

    void UDP::bind() throw (bind_error)
    {
      if ( m_local_interface.is_multicast() )
        {
          if (m_multicast_auto_join)
            {
              IPv4::Address addr( INADDR_ANY, m_local_interface.port() );
              Socket::bind(addr);
              join();
            }
          else
          {
            IPv4::Address addr( INADDR_ANY, 0 );
            Socket::bind(addr);
          }
        }
      else
        {
          IP::bind();
          if (m_local_interface.is_broadcast())
            set_option(SO_BROADCAST, true);
          else
            set_option(SO_BROADCAST, false);
        }
    }

    void UDP::bind(Conexus::Address& a) throw (bind_error)
    {
      IP::bind(a);
    }

    void UDP::close() throw (close_error)
    {
      if (m_multicast_request != NULL)
        leave();
      IP::close();
    }

    int UDP::pending_input_datagram_size()
    {
      int rsize;
      int error;
      error = ioctl(m_fd, SIOCINQ, &rsize);
      return rsize;
    }

    int UDP::pending_output_datagram_size()
    {
      int rsize;
      int error;
      error = ioctl(m_fd, SIOCOUTQ, &rsize);
      return rsize;
    }

    ssize_t UDP::write(const void* data, size_t size) throw (write_error)
    {
      if (m_write_without_connect)
        {
          if (!m_remote_address_set)
            throw error::write::no_default_remote_address();

          if ( is_closed() )
            try
              {
                change_state(OPENED);
              }
            catch (error::state::failed)
              {
                throw error::write::not_opened();
              }

          return Socket::writeto(m_remote_address, data, size);
        }
      else
        {
          if ( ! is_connected() )
            try
              {
                change_state(CONNECTED);
              }
            catch (error::state::failed)
              {
                throw error::write::not_connected();
              }

          return Socket::write(data, size);
        }
    }

    Data UDP::read(size_t s) throw (read_error)
    {
      if (! is_bound() )
        try
          {
            change_state(BOUND);
          }
        catch (error::state::failed)
          {
            throw error::read::not_bound();
          }

      return Socket::read(pending_input_datagram_size());
    }

    void UDP::connect( ) throw( connect_error )
    {
      if (m_remote_address_set)
        Socket::connect(m_remote_address);
      else
        throw error::connect::no_remote_address();
    }

    bool UDP::multicast_auto_join() const
      {
        return m_multicast_auto_join;
      }

    void UDP::set_multicast_auto_join(bool value)
    {
      m_multicast_auto_join = value;
    }

    bool UDP::write_without_connect() const
      {
        return m_write_without_connect;
      }


    void UDP::set_write_without_connect(bool value)
    {
      m_write_without_connect = value;
    }

    void UDP::join()
    {
      if ( m_multicast_request != NULL)
        leave();

      m_multicast_request = (struct ip_mreq*) calloc( 1, sizeof(struct ip_mreq) );
      m_multicast_request->imr_multiaddr.s_addr=m_local_interface.sockaddr_in().sin_addr.s_addr;
      m_multicast_request->imr_interface.s_addr=htonl(INADDR_ANY);
      set_option(SOL_IP, IP_ADD_MEMBERSHIP, *m_multicast_request);
    }

    void UDP::leave()
    {
      if ( m_multicast_request == NULL)
        return;

      set_option(SOL_IP, IP_DROP_MEMBERSHIP, *m_multicast_request);

      free(m_multicast_request);
      m_multicast_request = NULL;
    }

    int UDP::multicast_hops()
    {
      int i;
      option(SOL_IP, IP_MULTICAST_TTL, i);
      return i;
    }

    void UDP::set_multicast_hops(int i)
    {
      set_option(SOL_IP, IP_MULTICAST_TTL, i);
    }

    bool UDP::multicast_loop()
    {
      int i;
      option(SOL_IP, IP_MULTICAST_LOOP, i);
      return i;
    }

    void UDP::set_multicast_loop(bool b)
    {
      int i = b?1:0;
      set_option(SOL_IP, IP_MULTICAST_LOOP, i);
    }

    void UDP::on_local_interface_changed(unsigned which)
    {
      // if the socket is already bound, close and reopen to same state with new local port and interface
      if ( is_bound() )
        close_and_reopen();
    }

    void UDP::on_remote_address_changed(unsigned which)
    {
      // if the socket is already connected, close and reopen to same state with new remote port
      if ( is_connected() )
        close_and_reopen();
    }

  } // namespace IPv4
} // namespace conexus
