/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef CONEXUS_IPV4UDP_H
#define CONEXUS_IPV4UDP_H

#include <conexus/ipv4_ip.h>

/**
 * \page manual_udp Endpoints
 */

namespace Conexus
  {

  namespace IPv4
    {

    /**
     * This class encapsulates a socket used for UDP/IP communications.
     *
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     * @ingroup IPv4
     */
    class UDP : public IP
      {
      public:
        typedef boost::shared_ptr<UDP> pointer;
      
        UDP(uint16_t localport=0, uint16_t remoteport=0);
        
        static UDP::pointer create(uint16_t localport=0, uint16_t remoteport=0);

        virtual ~UDP() throw();

        virtual void bind() throw (bind_error);

        virtual void bind(Conexus::Address& a) throw (bind_error);

        virtual void close() throw (close_error);

        /**
         * Returns the size in bytes of the next pending datagram, or 0 when no datagram
         * is pending.
         */
        int pending_input_datagram_size();

        /**
         * Returns the size in bytes of the local send queue.
         */
        int pending_output_datagram_size();

        /**
         * Overloads the Socket ancestor send method to ensure that the socket is connected
         * before trying to send.
         */
        virtual ssize_t write(const void* data, size_t size) throw (write_error);

        /**
         * Receives a single datagram from the UDP socket up to s bytes in size; allocates
         * any memory necessary as a smart pointer contained in the returned Data object.
         *
         * If s < 0 then receives the next pending datagram, however large it may be.
         */
        virtual Data read(size_t s = -1)  throw (read_error);

        virtual void connect() throw (connect_error);
        virtual void connect(Address& a) throw (connect_error)
        {
          Socket::connect(a);
        }

        /**
         * There are some circumstances in which it may be desirable to set a
         * default address and have the send method transparently call the sendto method
         * without connecting rather than connecting and calling the send method.
         *
         * A specific example of this is when the remote address/port may occassionally
         * be unavailable. In this case the destination may generate an ICMP UNREACHABLE
         * message, and in accordance with RFC 1122 the local machine will fail any
         * future sends on a connected port, throwing a write::connection_refused error.
         * If you want to ensure that the datagram is sent contrary to RFC 1122 you could
         * set the default address, set send without connect and call send without throwing
         * connection errors.
         *
         * But, since this is contrary to RFC 1122 behavior you should really think about
         * your design before setting this.
         */

        void set_write_without_connect(bool value=true);


        bool write_without_connect() const;

        bool multicast_auto_join() const;

        void set_multicast_auto_join(bool value=true);

        /**
         * Returns the multicast ttl of the currently associated interface.
         */
        int multicast_hops();

        /**
         * Sets the multicast ttl of the currently associated interface to @em hops.
         */
        void set_multicast_hops(int hops);

        /**
         * If true multicast packets are looped back to local sockets.
         */
        bool multicast_loop();

        /**
         * If parameter @em loop is true then multicast packets are looped back to local sockets.
         */
        void set_multicast_loop(bool loop);

        virtual const std::string& object_type() { static std::string s("Conexus::IPv4::UDP"); return s; }

        protected:
        bool m_write_without_connect;
        bool m_multicast_auto_join;
        struct ip_mreq* m_multicast_request;


        /**
         * Joins the multicast group specified by @em addr.
         *
         * @param ifindex is  the  interface index of the interface
         *    that should join the multicast group, or 0 to indicate
         *    any interface.
         */
        void join();

        /**
         * Leaves the multicast group specified by @em addr.
         *
         * @param ifindex is  the  interface index of the interface
         *    that should leave the multicast group, or 0 to indicate
         *    any interface.
         */
        void leave();

        virtual void on_local_interface_changed(unsigned which);
        virtual void on_remote_address_changed(unsigned which);
      };

  }

}

#endif
