/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#ifndef CONEXUS_IPV6TCP_H
#define CONEXUS_IPV6TCP_H

#include <boost/shared_ptr.hpp>

#include <conexus/ipv6_ip.h>

/**
 * \page manual_udp Endpoints
 */

namespace Conexus
  {

  namespace IPv6
    {

    /**
     * This class encapsulates a socket used for TCP/IP communications.
     *
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     * @ingroup IPv6
     */
    class TCP : public IP
      {
      public:
        friend class TCPServer;

        typedef boost::shared_ptr<TCP> pointer;

        TCP(uint16_t port=0);

        static TCP::pointer create(uint16_t port=0);

        virtual ~TCP() throw();

        /**
         * Returns the size in bytes of the next pending datagram, or 0 when no datagram
         * is pending.
         */
        int pending_input_size();

        /**
         * Returns the size in bytes of the local send queue.
         */
        int pending_output_size();

        /**
         * Overloads the Socket ancestor send method to ensure that the socket is connected
         * before trying to send.
         */
        virtual ssize_t write(const void* data, size_t size) throw (write_error);

        /**
         * Receives a single datagram from the TCP socket up to s bytes in size; allocates
         * any memory necessary as a smart pointer contained in the returned Data object.
         *
         * If s < 0 then receives the next pending datagram, however large it may be.
         */
        virtual Data read(size_t s = -1)  throw (read_error);

        virtual void connect() throw (connect_error);
        virtual void connect(Address& a) throw (connect_error)
        {
          Socket::connect(a);
        }
        
        virtual const std::string& object_type() { static std::string s("Conexus::IPv6::TCP"); return s; }

      protected:

        /**
         * Protected constructor to be used by TCPServer to create TCP connections.
         * The second bool parameter is used to differentiate the public method from the private method.
         */
        TCP(int accepted_fd, bool);

        static TCP::pointer create(int accepted_fd, bool);

        virtual void on_local_interface_changed(unsigned which);
        virtual void on_remote_address_changed(unsigned which);
      };

  }

}

#endif
