/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "rwfiledescriptor.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

namespace Conexus {

RWFileDescriptor::RWFileDescriptor(unsigned long readwrite):
  FileDescriptor()
{
  change_state(readwrite&(READ|WRITE));
}

RWFileDescriptor::~RWFileDescriptor()
{
}

ssize_t RWFileDescriptor::write( const void * data, size_t size, IOMETHOD block ) throw (write_error)
{
  if (is_write_blocked())
    return 0;

  if ( is_closed() )
    try {
    change_state(OPENED|(state()&(READ|WRITE)));
    } catch (error::state::failed) {
      throw error::write::not_opened();
    }

  return ::write(m_fd, data, size);
  // TODO come back and check error codes
}

Data RWFileDescriptor::read( size_t s ) throw (read_error)
{
  Data d;
  if (is_read_blocked())
    return d;

  if ( is_closed() )
    try {
    change_state(OPENED|(state()&(READ|WRITE)));
    } catch (error::state::failed) {
      throw error::read::not_opened();
    }

  d.data = Data::Octets(new Data::Octet[s]);
  d.size = ::read(m_fd, d.data.get(), s);
  // TODO come back and check error codes
  return d;
}

void RWFileDescriptor::change_state( long unsigned new_state ) throw (state_error)
{
  if (new_state == UNCHANGED)
    return;

  // if the READ or WRITE flags changed at a minimum we need to close the port to
  // force a reopen
  if ( m_state&(READ|WRITE) != new_state&(READ|WRITE) )
    Endpoint::change_state(CLOSED);

  // change the READ and WRITE flags
  // zero out the READWRITE bits, then bitwise-or in only those two bits
  set_state_readable(new_state&READ);
  set_state_writable(new_state&WRITE);

  // check to see if the port needs to be opened
  Endpoint::change_state(new_state);
}

      sigc::signal<void,bool> RWFileDescriptor::signal_readable_changed()
{
  return m_signal_readable_changed;
}
      
      sigc::signal<void,bool> RWFileDescriptor::signal_writable_changed()
{
  return m_signal_writable_changed;
}

      bool RWFileDescriptor::is_readable()
{
  return m_state&READ;
}

      bool RWFileDescriptor::is_writable()
{
  return m_state&WRITE;
}

void RWFileDescriptor::set_state_readable(bool b) {
  if (is_readable() && b)
    return;
  if (!is_readable() && !b)
    return;

  if (b)
    m_state |= READ;
  else
    m_state &= ~READ;
  m_signal_readable_changed.emit(b);
}

void RWFileDescriptor::set_state_writable(bool b) {
  if (is_writable() && b)
    return;
  if (!is_writable() && !b)
    return;

  if (b)
    m_state |= WRITE;
  else
    m_state &= ~WRITE;
  m_signal_writable_changed.emit(b);
}



}
