/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef CONEXUSSERIAL_H
#define CONEXUSSERIAL_H

#include <conexus/rwfiledescriptor.h>

/**
 * \page manual_serial Serial Device Endpoints
 */

namespace Conexus
  {

  /**
   * This class encapsulates a raw serial port. It is likely that it will be extended one
   * level deeper in the inheritance hierarchy and replaced by an ancestor that provides
   * file I/O to be used with traditional files and with Unix named pipes.
   *
   * @author Rick L Vinyard Jr
   * @ingroup conexus
   */
  class Serial : public RWFileDescriptor
    {
    public:
      typedef boost::shared_ptr<Serial> pointer;
      
      Serial(const std::string& device=std::string(), unsigned long readwrite=READ|WRITE);
      
      static Serial::pointer create(const std::string& device=std::string(), unsigned long readwrite=READ|WRITE);

      virtual ~Serial();

      /**
       * Opens the named serial port device in the specified mode.
       */
      virtual void open(const std::string name, int state=READ|WRITE) throw (open_error);

      /**
       * Will attempt to open the serial port if a device has been specified. Otherwise an
       * error will be thrown.
       */
      virtual void open() throw (open_error);

      /**
       * Will return the string representing the device associated with this serial object.
       */
      std::string device();

      /**
       * Sets the string representing the device associated with this serial object.
       *
       * If the serial device is currently open and d is not equal to the current device,
       * the current device will be closed.
       *
       * Also, if auto_open is set, the new device will be immediately opened.
       *
       * Additionally, this method checks whether the parameter d is the same as the current device,
       * and parameter mode is the same as the current mode.
       * If both are equal, the device will not be closed and reopened unless force=true.
       */
      void set_device(const std::string& d, long unsigned state=READ|WRITE) throw (open_error);

      sigc::signal<void> signal_device_changed();
      
      virtual void start(ReadWrite rw=Conexus::READ, bool use_dispatcher=false);

      virtual const std::string& object_type() { static std::string s("Conexus::Serial"); return s; }

    protected:
      std::string m_device;

      sigc::signal<void> m_signal_device_changed;
    };

}

#endif
