/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef CONEXUSSERVER_H
#define CONEXUSSERVER_H

#include <set>
#include <queue>

#include <sys/select.h>
#include <pthread.h>

#include <sigc++/sigc++.h>

#include <glibmm/dispatcher.h>
#include <glibmm/thread.h>

#include <conexus/object.h>
#include <conexus/data.h>

namespace Conexus
  {

  /**
  @author Rick L Vinyard Jr
   * @ingroup conexus
   */
  class Server: public Object
    {
    public:
      Server();
      virtual ~Server();

      virtual void start(bool use_dispatcher=false);
      virtual void stop();

      /**
       * Sets the responsiveness to add_*, remove_* and stop commands in microseconds;
       * lower is more responsive; larger is less responsive; default is 5000 microseconds (or 5 ms).
       *
       * In general this refers to the timeout period of the @em select() call used internally.
       * Each cycle of the server I/O thread acquires several mutex locks at the beginning
       * of the cycle and releases them at the end of the cycle.
       *
       * This value shortens the cycle period, thereby making the server more responsive to commands at
       * the cost of increasing overhead.
         */
      virtual void set_responsiveness(long r) = 0;

      /**
       * Returns the server responsiveness level.
       */
      virtual long responsiveness() = 0;

      virtual bool is_running();

      virtual const std::string& object_type() { static std::string s("Conexus::Server"); return s; }

    protected:
      pthread_t m_thread;
      pthread_attr_t m_thread_attr;
      bool m_terminate;
      bool m_running;
      Glib::Dispatcher* m_dispatcher;
      bool m_use_dispatcher;


      virtual void service_thread_main() = 0;

    private:
      static void* service_thread_main_static_proxy(void*);
      void service_thread_main_proxy();

    };

}

#endif
