/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef CONEXUSTTY_H
#define CONEXUSTTY_H

#include <conexus/serial.h>

#include <termios.h>
#include <unistd.h>

/**
 * \page manual_tty Serial TTY Endpoints
 */

namespace Conexus
  {

  /**
   * This class extends Serial to include operations upon TTY (TeleTYpe) serial
   * communications. This class should be used to control serial port baud rates
   * and other aspects of serial communications.
   *
   * @author Rick L Vinyard Jr
   * @ingroup conexus
   */
  class TTY : public Serial
    {
    public:
      typedef boost::shared_ptr<TTY> pointer;
      
      /**
       * Default constructor
       * The following default characteristics are set:
       *   - Byte size: 8 bits
       *   - Receiver: Enabled
       *   - Hardware Flow Control: Enabled
       *   - Hardware Carrier Detect: Disabled (modem status lines ignored)
       *   - Parity: None
       *   - Breaks: Ignored
       *   - Delays set to 0: newline, carriage return, horizontal tab, backspace, vertical tab, form feed
       */
      TTY(const std::string& device=std::string());

      static TTY::pointer create(const std::string& device=std::string());
      
      virtual ~TTY();

      /**
       * Aliases for the tty function options relating to when an option change takes effect.
       */
      typedef enum SetOption {
        /** the setting change occurs immediately */
        SET_NOW=TCSANOW,

        /**
         * the setting change occurs after all output written to the object
         * has been transmitted, and all input that has been received but not
         * read will be discarded before the change is made.
        */
        SET_FLUSH=TCSAFLUSH,

        /**
         * the setting change occurs after all output has been transmitted;
         * This function should be used when changing parameters that affect output.
         */
        SET_DRAIN=TCSADRAIN
      } SetOption;

      typedef enum Parity {
        PARITY_NONE, /** No parity check */
        PARITY_ODD,  /** Odd parity checks */
        PARITY_EVEN, /** Even parity checks */
      };
      
      typedef enum ParityError {
        PARITY_ERROR_DISCARD, /** Discards all parity error bytes */
        PARITY_ERROR_MARK,    /** Delivers three bytes (0xFF, 0x00 and c; where c is the error byte */
        PARITY_ERROR_ZERO,    /** Delivers 0x00 when a parity error is detected */
        PARITY_ERROR_IGNORE,  /** Ignores all parity errors and delivers all data as is (this is not the same as IGNPAR) */
      };
      
      typedef enum FlowControl {
        FLOW_NONE, /** CRTSCTS off, IXON off, IXOFF off */
        FLOW_HARDWARE, /** CRTSCTS on, IXON off, IXOFF off */
        FLOW_SOFTWARE, /** CRTSCTS off, IXON on, IXOFF on */
        FLOW_CUSTOM, /** Combinations of CRTSCTS, IXON, IXOFF other than FLOW_NONE, FLOW_HARDWARE, or FLOW_SOFTWARE */
      };

      /**
       * waits until all output written to the object has been transmitted.
       */
      void drain();

      /**
       * discards data received but not read
       */
      void flush_input();

      /**
       * discards data written to the object but not transmitted
       */
      void flush_output();

      /**
       * discards both data written to the object but not transmitted and data
       * received but not read
       */
      void flush();

      /**
       * Suspends transmission of data by the object
       */
      void suspend_output();

      /**
       * Resumes transmission of data by the object
       */
      void restart_output();

      /**
       * Suspends reception of data by the object
       */
      void suspend_input();

      /**
       * Resumes reception of data by the object
       */
      void restart_input();

      /**
       * Sets the input line speed
       * 
       * These are generally defined in /usr/include/bits/termios.h
       * The following is the recognized list:
       * - 0 (hang up)
       * - 50
       * - 75
       * - 110
       * - 134
       * - 150
       * - 200
       * - 300
       * - 600
       * - 1200
       * - 1800
       * - 2400
       * - 4800
       * - 9600
       * - 19200
       * - 38400
       * - 57600
       * - 115200
       * - 230400
       * - 460800
       * - 500000
       * - 576000
       * - 921600
       * - 1000000
       * - 1152000
       * - 1500000
       * - 2000000
       * - 2500000
       * - 3000000
       * - 3500000
       * - 4000000
       */
      void set_input_speed(unsigned speed, SetOption option=SET_NOW );

      /** Sets the output line speed */
      void set_output_speed(unsigned speed, SetOption option=SET_NOW );

      /** Sets both the input and output line speeds */
      void set_speed(unsigned speed, SetOption option=SET_NOW );

      /** Returns the current input line speed */
      unsigned input_speed();

      /** Returns the current output line speed */
      unsigned output_speed();

      /** Sets parity mode */
      void set_parity(Parity parity, ParityError error=PARITY_ERROR_IGNORE, SetOption option=SET_NOW );
      
      /** Returns the current parity mode */
      Parity parity();
      
      /** Returns the current parity error mode */
      ParityError parity_error();

      /** Get the current byte size in bits */
      unsigned byte_size();

      /**
       * Sets the communication byte size
       * Valid values are:
       *  - 5 bits-per-byte
       *  - 6 bits-per-byte
       *  - 7 bits-per-byte
       *  - 8 bits-per-byte (default)
       * All other values are ignored.
       */
      void set_byte_size( unsigned size, SetOption option=SET_NOW );

      /** Get the number of stop bits (1 or 2) */
      unsigned stop_bits();
      
      /** Sets the stop bits; only values of 1 or 2 will be processed, all others will be ignored. */
      void set_stop_bits( unsigned size, SetOption option=SET_NOW );
      
      /** Returns the flow control method. */
      FlowControl flow_control();

      /**
       * Sets the flow control flags according to the flow control method specified.
       * If FLOW_CUSTOM is specified, no change will occur.
       */
      void set_flow_control(FlowControl flowcontrol, SetOption option=SET_NOW);

      /** True if the hardware carrier detect (DCD) is enabled, false otherwise. */
      bool carrier_detect_enabled();
      
      /**
       * Enables or disables a hardware carrier detect.
       * If hardware carrier detect is enabled and this is a blocking TTY, an open will 
       * block until a DCD signal is received from the hardware (useful for modems).
       * If this is not enabled, the TTY will behave as if the DCD is always asserted.
       * The hardware carrier flag should usually be turned on for modems and off for
       * local devices.
       */
      void set_carrier_detect(bool enable=true, SetOption option=SET_NOW);
      
      /**
       * True if the receive side is enabled, false otherwise.
       * If false, this TTY cannot receive characters.
       */
      bool receiver_enabled();

      /**
       * If true the receiver will be enabled, and if false it will be disabled
       * and no characters can be received.
       */
      void set_receiver_enabled(bool enable=true, SetOption option=SET_NOW);

      /**
       * Reads up to s bytes from the TTY
       */
      virtual Data read(size_t s = 0) throw (read_error);

      virtual void open() throw (open_error);

      /** Opens the TTY in the specified mode, storing the original terminal mode, to be restored on close. */
      virtual void open(const std::string name, int s=READ|WRITE) throw (open_error);

      /** Restores the original terminal modes on the tty */
      virtual void close(bool force=false) throw (close_error);

      /**
       * Sets TTY modes [ see termios(3) ]
       */
      void set_input_modes(tcflag_t iflag, SetOption option=SET_NOW);

      /**
       * Sets TTY modes [ see termios(3) ]
       */
      void set_output_modes(tcflag_t oflag, SetOption option=SET_NOW);

      /**
       * Sets TTY modes [ see termios(3) ]
       */
      void set_control_modes(tcflag_t cflag, SetOption option=SET_NOW);

      /**
       * Sets TTY modes [ see termios(3) ]
       */
      void set_local_modes(tcflag_t lflag, SetOption option=SET_NOW);

      /**
       * Sets TTY modes [ see termios(3) ]
       */
      void set_control_characters(int index, cc_t value, SetOption option=SET_NOW);

      /**
       * Gets TTY attributes [ see termios(3) ]
       */
      struct termios attributes();
      
      /**
       * Injects the character into the input queue
       */
      void inject(char data);
      
      /**
       * Injects the input buffer into the input queue
       */
      void inject(const char* buffer, size_t bufsize);

      /**
       * Injects the input buffer into the input queue
       */
      void inject(const std::string& buffer);

      /** True if the terminal will be reset to original characteristics on close. */
      bool reset_on_close();
      
      /** Set to true if the terminal should reset to original characteristics on close. */
      void set_reset_on_close(bool reset=true);
      
      sigc::signal<void> signal_input_speed();
      sigc::signal<void> signal_output_speed();
      sigc::signal<void> signal_parity();
      sigc::signal<void> signal_byte_size();
      sigc::signal<void> signal_stop_bits();
      sigc::signal<void> signal_flow_control();
      sigc::signal<void> signal_carrier_detect();
      sigc::signal<void> signal_receiver();

      virtual const std::string& object_type() { static std::string s("Conexus::TTY"); return s; }

      protected:
      struct termios m_termios, m_origtermios;
      bool m_reset_on_close;

      sigc::signal<void> m_signal_input_speed;
      sigc::signal<void> m_signal_output_speed;
      sigc::signal<void> m_signal_parity;
      sigc::signal<void> m_signal_byte_size;
      sigc::signal<void> m_signal_stop_bits;
      sigc::signal<void> m_signal_flow_control;
      sigc::signal<void> m_signal_carrier_detect;
      sigc::signal<void> m_signal_receiver;

      void tcsetattr(SetOption option);

      speed_t unsigned2speed(unsigned speed);
      unsigned speed2unsigned(speed_t speed);

    };

}

#endif
